@@ -8,16 +8,17 @@ from pathlib import Path
 from olefile import OleFileIO, isOleFile
 import numpy as np
 import tifffile as tf
-from oxdls import OMEXML
+from oxdls import OMEXML, DO_XYZCT, PT_UINT16, PT_UINT16, PT_FLOAT, PT_DOUBLE
 
 from . import txrm_wrapper
 from .annotator import Annotator
 
+from .__init__ import __version__
 
 dtype_dict = {
-    'uint16': np.uint16,
-    'float32': np.float32,
-    'float64': np.float64
+    "uint16": PT_UINT16,
+    "float32": PT_FLOAT,
+    "float64": PT_DOUBLE
 }
 
 def _dynamic_despeckle_and_average_series(images, average=True):
@@ -135,27 +136,38 @@ def manual_save(tiff_file, image, data_type=None, metadata=None):
     image = _cast_to_dtype(image, data_type)
     if metadata is not None:
         meta_img = metadata.image()
-        meta_img.Pixels.set_PixelType(str(image.dtype))
-        meta_img.set_Name(tiff_path.name)
+        meta_img.Pixels.set_PixelType(dtype_dict[str(image.dtype)])
+        meta_img.set_Name(tiff_path.stem)
         metadata = metadata.to_xml().encode()
 
     tiff_dir = tiff_path.resolve().parent
     tiff_dir.mkdir(parents=True, exist_ok=True)
     num_frames = len(image)
+    bigtiff = image.size * image.itemsize >= np.iinfo(np.uint32).max  # Check if data bigger than 4GB TIFF limit
+
     logging.info("Saving image as %s with %i frames", tiff_path.name, num_frames)
 
-    with tf.TiffWriter(str(tiff_path), bigtiff=True) as tif:
-        tif.save(image[0], photometric='minisblack', description=metadata, metadata={'axes':'XYZCT'})
-        for i in range(1, num_frames):
-            tif.save(image[i], photometric='minisblack', metadata={'axes':'XYZCT'})
-
+    with tf.TiffWriter(str(tiff_path), bigtiff=bigtiff, ome=False) as tif:
+        tif.write(
+            image,
+            photometric='MINISBLACK',
+            description=metadata,
+            metadata={'axes': 'ZYX'},
+            software=f"txrm2tiff {__version__}"
+            )
+    
 def save_colour(tiff_file, image):
     tiff_path = Path(tiff_file)
     num_frames = len(image)
-    logging.info("Saving image as %s with %i frames", tiff_path.name, num_frames)
-    with tf.TiffWriter(str(tiff_path), bigtiff=True) as tif:
-        for i in range(0, num_frames):
-            tif.save(image[i], photometric='rgb')
+    logging.info("Saving annotated image as %s with %i frames", tiff_path.name, num_frames)
+    bigtiff = image.size * image.itemsize >= np.iinfo(np.uint32).max  # Check if data bigger than 4GB TIFF limit
+    with tf.TiffWriter(str(tiff_path), bigtiff=bigtiff, ome=False) as tif:
+        tif.write(
+            image,
+            photometric='RGB',
+            metadata={'axes':'ZYXC'},
+            software=f"txrm2tiff {__version__}"
+            )
 
 
 def _convert_output_path_to_annotated_path(output_path):
@@ -171,13 +183,31 @@ def _convert_output_path_to_annotated_path(output_path):
         suffix = f".{split_name[-1]}"
         name_idx -= 1
         if num_parts > 2 and split_name[-2].lower() == "ome":
-            suffix = f".{split_name[-2]}{suffix}"
             name_idx -= 1
     annotated_name = ".".join([s for s in split_name[:name_idx]])
     annotated_name += f"_Annotated{suffix}"
     return output_path.parent / annotated_name
 
 
+def normalise_to_datatype(array, datatype, clip=False):
+    logging.debug("Re-scaling array to %s", datatype)
+    array = array.astype(np.float64)  # Convert to float for rescaling
+
+    # Clip outliers before normalisation to avoid any speckles
+    # that can lead to the normalised image to be dark:
+    if clip:
+        num_std = 3.
+        logging.info("Clipping outliers (>%g std from mean)", num_std)
+        new_max = np.mean(array) + num_std * np.std(array)
+        np.clip(array, 0, new_max, out=array)
+
+    # Move minimum value of all corrected images to 0:
+    array -= array.min()
+    # New max should be the max allowed by datatype
+    new_max = np.iinfo(datatype).max
+    return array * (new_max / array.max())
+
+
 def _cast_to_dtype(image, data_type):
     if data_type is not None:
         try:
@@ -190,12 +220,11 @@ def _cast_to_dtype(image, data_type):
                     logging.warning(
                         "Image min %f below %s minimum of %i, values below this will be cut off",
                         img_min, dtype, dtype_info.min)
-                    _conditional_replace(image, dtype_info.min, lambda x: x < dtype_info.min)
                 if dtype_info.max < img_max:
                     logging.warning(
                         "Image max %f above %s maximum of %i, values above this will be cut off",
                         img_max, dtype, dtype_info.max)
-                    _conditional_replace(image, dtype_info.max, lambda x: x > dtype_info.max)
+                np.clip(image, dtype_info.min, dtype_info.max, out=image)
             return np.around(image, decimals=0).astype(dtype)
         except Exception:
             logging.error("Invalid data type given: %s aka %s. Saving with default data type.", data_type, dtype)
@@ -208,7 +237,6 @@ def create_ome_metadata(ole, image_list, filename=None):
     # Get image dimensions
     # X, Y, number of frames (T in tilt series):
     dimensions = (*image_list[0].shape, len(image_list))
-    str_dtype = str(image_list[0].dtype)
 
     # Get metadata variables from ole file:
     exposures = txrm_wrapper.extract_multiple_exposure_times(ole)
@@ -226,27 +254,32 @@ def create_ome_metadata(ole, image_list, filename=None):
     ox = OMEXML()
 
     image = ox.image()
-    image.set_ID("0")
+    image.set_ID("Image:0")
     image.set_AcquisitionDate(date_time)
     if filename:
         image.set_Name(filename)
 
     pixels = image.Pixels
-    pixels.set_DimensionOrder("XYTZC")
-    pixels.set_ID("0")
-    pixels.set_PixelType(str_dtype)
+    pixels.DimensionOrder = DO_XYZCT
+    pixels.set_ID("Pixels:0")
+    
     pixels.set_SizeX(dimensions[0])
     pixels.set_SizeY(dimensions[1])
-    pixels.set_SizeT(dimensions[2])
-    pixels.set_SizeZ(1)
     pixels.set_SizeC(1)
+    pixels.set_SizeZ(dimensions[2])
+    
+    pixels.set_SizeT(1)
+
     pixels.set_PhysicalSizeX(physical_img_sizes[0])
     pixels.set_PhysicalSizeXUnit("nm")
     pixels.set_PhysicalSizeY(physical_img_sizes[1])
     pixels.set_PhysicalSizeYUnit("nm")
+    pixels.set_PhysicalSizeZ(1)
+    pixels.set_PhysicalSizeZUnit("reference frame")
 
+    pixels.set_tiffdata_count(dimensions[2])  # tiffdata must be created before planes
     pixels.set_plane_count(dimensions[2])
-    pixels.set_tiffdata_count(dimensions[2])
+    
 
     channel = pixels.Channel(0)
     channel.set_ID("Channel:0:0")
@@ -304,6 +337,13 @@ def create_ome_metadata(ole, image_list, filename=None):
 
     # Add plane/tiffdata for each plane in the stack
     for count in range(dimensions[2]):
+        tiffdata = pixels.tiffdata(count)
+        tiffdata.set_FirstC(0)
+        tiffdata.set_FirstZ(count)
+        tiffdata.set_FirstT(0)
+        tiffdata.set_IFD(count)
+        tiffdata.set_plane_count(1)
+
         plane = pixels.Plane(count)
         plane.set_ExposureTime(exposures[count])
         plane.set_TheZ(count)
@@ -317,13 +357,6 @@ def create_ome_metadata(ole, image_list, filename=None):
         plane.set_PositionY(y_positions[count])
         plane.set_PositionZ(count)
 
-        tiffdata = pixels.tiffdata(count)
-        tiffdata.set_FirstC(0)
-        tiffdata.set_FirstZ(count)
-        tiffdata.set_FirstT(0)
-        tiffdata.set_IFD(count)
-        tiffdata.set_plane_count(1)
-
     return ox
 
 class TxrmToImage:
@@ -357,7 +390,11 @@ class TxrmToImage:
                 else:
                     self.annotator = False
             # Create metadata
-            self.ome_metadata = create_ome_metadata(ole, self.image_output)
+            try:
+                self.ome_metadata = create_ome_metadata(ole, self.image_output)
+            except Exception:
+                logging.error("Error occurred creating metadata", exc_info=True)
+                self.ome_metadata = None
 
     def get_image_and_metadata(self):
         if self.image_output is None:
@@ -369,7 +406,10 @@ class TxrmToImage:
 
     def get_annotated_images(self):
         if self.annotator is not None:
-            return self.annotator.apply_annotations(self.image_output)
+            logging.info("Creating a normalised image to apply annotations to")
+            return self.annotator.apply_annotations(
+                normalise_to_datatype(self.image_output, np.uint8, True)
+            )
         return None
 
     def save(self, tiff_file, data_type=None):
@@ -381,7 +421,7 @@ class TxrmToImage:
             elif not self.annotator:
                 logging.error("No annotations or scale bar to save with %s", tiff_path)
             else:
-                annotationed_images = self.annotator.apply_annotations(self.image_output)
+                annotationed_images = self.get_annotated_images()
                 if annotationed_images is not None:
                     annotated_path = _convert_output_path_to_annotated_path(tiff_path)
                     save_colour(annotated_path, annotationed_images)
